# automatically generated by the FlatBuffers compiler, do not modify

# namespace: meta

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class Reaction(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Reaction()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsReaction(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Reaction
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Table of the object holding the attribute.
    # Reaction
    def TableOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Reaction
    def TableOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Reaction
    def TableOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Reaction
    def TableOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Object (within the table) holding the attribute
    # Reaction
    def ObjectOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Reaction
    def ObjectOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Reaction
    def ObjectOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Reaction
    def ObjectOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Object (within the table) holding the attribute
    # Reaction
    def VoterOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Reaction
    def VoterOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Reaction
    def VoterOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Reaction
    def VoterOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # Reaction name (or URI in general).
    # Reaction
    def Reaction(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Timestamp when the reaction was recorded or last modifified.
    # Reaction
    def Reacted(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # CBOR-serialized, object-valued extra data stored along with this reaction.
    # Reaction
    def Extra(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Reaction
    def ExtraAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Reaction
    def ExtraLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Reaction
    def ExtraIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0


def ReactionStart(builder):
    builder.StartObject(6)


def Start(builder):
    ReactionStart(builder)


def ReactionAddTableOid(builder, tableOid):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(tableOid), 0)


def AddTableOid(builder, tableOid):
    ReactionAddTableOid(builder, tableOid)


def ReactionStartTableOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartTableOidVector(builder, numElems: int) -> int:
    return ReactionStartTableOidVector(builder, numElems)


def ReactionAddObjectOid(builder, objectOid):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(objectOid), 0)


def AddObjectOid(builder, objectOid):
    ReactionAddObjectOid(builder, objectOid)


def ReactionStartObjectOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartObjectOidVector(builder, numElems: int) -> int:
    return ReactionStartObjectOidVector(builder, numElems)


def ReactionAddVoterOid(builder, voterOid):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(voterOid), 0)


def AddVoterOid(builder, voterOid):
    ReactionAddVoterOid(builder, voterOid)


def ReactionStartVoterOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartVoterOidVector(builder, numElems: int) -> int:
    return ReactionStartVoterOidVector(builder, numElems)


def ReactionAddReaction(builder, reaction):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(reaction), 0)


def AddReaction(builder, reaction):
    ReactionAddReaction(builder, reaction)


def ReactionAddReacted(builder, reacted):
    builder.PrependUint64Slot(4, reacted, 0)


def AddReacted(builder, reacted):
    ReactionAddReacted(builder, reacted)


def ReactionAddExtra(builder, extra):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(extra), 0)


def AddExtra(builder, extra):
    ReactionAddExtra(builder, extra)


def ReactionStartExtraVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartExtraVector(builder, numElems: int) -> int:
    return ReactionStartExtraVector(builder, numElems)


def ReactionEnd(builder):
    return builder.EndObject()


def End(builder):
    return ReactionEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
from typing import Optional
from uuid import UUID

from cfxdb.mrealm.cluster_node_membership import ClusterNodeMembership


class RouterClusterNodeMembership(ClusterNodeMembership):
    """
    Membership of a managed node (paired within a management realm) to a router cluster.
    A router cluster is able to run "data planes", which are groups of router workers kept in sync,
    and meshed via router-to-router links. Finally, "(application) realms" can be started in data planes.
    """

    def __init__(
        self,
        cluster_oid: Optional[UUID] = None,
        node_oid: Optional[UUID] = None,
        softlimit: Optional[int] = None,
        hardlimit: Optional[int] = None,
        _unknown=None,
    ):
        """

        :param cluster_oid: Object ID of router cluster the node is becoming member of.

        :param node_oid: Object ID of the node becoming member of the router cluster.

        :param softlimit: Limits the number of router workers started on this node. When this number of workers
            is reached, efforts are taken to actively reduce the resource load on this node (eg stopping
            other running router workers "currently being idle"). New router workers will still be started.

        :param hardlimit: Maximum number of router workers started on this node. This hard limits the
            number of router workers started on this node.
        """
        ClusterNodeMembership.__init__(self, cluster_oid=cluster_oid, node_oid=node_oid, _unknown=_unknown)
        self.softlimit = softlimit
        self.hardlimit = hardlimit

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if not ClusterNodeMembership.__eq__(self, other):
            return False
        if other.softlimit != self.softlimit:
            return False
        if other.hardlimit != self.hardlimit:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return pprint.pformat(self.marshal())

    def marshal(self):
        """
        Marshal this object to a generic host language object.

        :return: dict
        """
        obj = ClusterNodeMembership.marshal(self)
        obj.update(
            {
                "softlimit": self.softlimit,
                "hardlimit": self.hardlimit,
            }
        )
        return obj

    @staticmethod
    def parse(data):
        """
        Parse generic host language object into an object of this class.

        :param data: Generic host language object
        :type data: dict

        :return: instance of :class:`WebService`
        """
        assert type(data) == dict

        obj = ClusterNodeMembership.parse(data)
        data = obj._unknown or {}

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = {}
        for k in data:
            if k not in ["softlimit", "hardlimit"]:
                _unknown[k] = data[k]

        softlimit = None
        if "softlimit" in data and data["softlimit"]:
            assert type(data["softlimit"]) == int
            softlimit = data["softlimit"]

        hardlimit = None
        if "hardlimit" in data and data["hardlimit"]:
            assert type(data["hardlimit"]) == int
            hardlimit = data["hardlimit"]

        obj = RouterClusterNodeMembership(
            cluster_oid=obj.cluster_oid,
            node_oid=obj.node_oid,
            softlimit=softlimit,
            hardlimit=hardlimit,
            _unknown=_unknown,
        )

        return obj



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: meta

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# Generic **object documentation** attached to objects in other tables. Primary key of this table is ``oid``.
class Doc(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Doc()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsDoc(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Doc
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # OID of the documentation (version).
    # Doc
    def Oid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Table of the object being documented.
    # Doc
    def TableOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Doc
    def TableOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Doc
    def TableOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Doc
    def TableOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Object (within the table)  being documented.
    # Doc
    def ObjectOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Doc
    def ObjectOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Doc
    def ObjectOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Doc
    def ObjectOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # When the documentation (version) was created or modified.
    # Doc
    def Modified(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Documentation text format.
    # Doc
    def Format(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # The actual documentation, serialized according to the documentation format.
    # Doc
    def Document(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Doc
    def DocumentAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Doc
    def DocumentLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Doc
    def DocumentIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0


def DocStart(builder):
    builder.StartObject(6)


def Start(builder):
    DocStart(builder)


def DocAddOid(builder, oid):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(oid), 0)


def AddOid(builder, oid):
    DocAddOid(builder, oid)


def DocAddTableOid(builder, tableOid):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(tableOid), 0)


def AddTableOid(builder, tableOid):
    DocAddTableOid(builder, tableOid)


def DocStartTableOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartTableOidVector(builder, numElems: int) -> int:
    return DocStartTableOidVector(builder, numElems)


def DocAddObjectOid(builder, objectOid):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(objectOid), 0)


def AddObjectOid(builder, objectOid):
    DocAddObjectOid(builder, objectOid)


def DocStartObjectOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartObjectOidVector(builder, numElems: int) -> int:
    return DocStartObjectOidVector(builder, numElems)


def DocAddModified(builder, modified):
    builder.PrependUint64Slot(3, modified, 0)


def AddModified(builder, modified):
    DocAddModified(builder, modified)


def DocAddFormat(builder, format):
    builder.PrependUint8Slot(4, format, 0)


def AddFormat(builder, format):
    DocAddFormat(builder, format)


def DocAddDocument(builder, document):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(document), 0)


def AddDocument(builder, document):
    DocAddDocument(builder, document)


def DocStartDocumentVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartDocumentVector(builder, numElems: int) -> int:
    return DocStartDocumentVector(builder, numElems)


def DocEnd(builder):
    return builder.EndObject()


def End(builder):
    return DocEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from typing import Optional

import zlmdb

from cfxdb.realmstore._event import Events
from cfxdb.realmstore._publication import Publications
from cfxdb.realmstore._session import IndexSessionsBySessionId, Sessions


class RealmStore(object):
    """
    Persistent realm store.
    """

    __slots__ = (
        "_db",
        "_sessions",
        "_idx_sessions_by_session_id",
        "_publications",
        "_events",
    )

    def __init__(self, db):
        self._db = db
        self._sessions: Optional[Sessions] = None
        self._idx_sessions_by_session_id: Optional[IndexSessionsBySessionId] = None
        self._publications: Optional[Publications] = None
        self._events: Optional[Events] = None

    @property
    def db(self) -> zlmdb.Database:
        """
        Database this schema is attached to.
        """
        return self._db

    @property
    def sessions(self) -> Optional[Sessions]:
        """
        Sessions persisted in this realm store.
        """
        return self._sessions

    @property
    def idx_sessions_by_session_id(self) -> Optional[IndexSessionsBySessionId]:
        """
        Index: (session, joined_at) -> app_session_oid
        """
        return self._idx_sessions_by_session_id

    @property
    def publications(self) -> Optional[Publications]:
        """
        Publications archive.
        """
        return self._publications

    @property
    def events(self) -> Optional[Events]:
        """
        Events archive.
        """
        return self._events

    @staticmethod
    def attach(db: zlmdb.Database) -> "RealmStore":
        schema = RealmStore(db)

        schema._sessions = db.attach_table(Sessions)

        assert schema._sessions is not None

        schema._idx_sessions_by_session_id = db.attach_table(IndexSessionsBySessionId)
        schema._sessions.attach_index(
            "idx1", schema._idx_sessions_by_session_id, lambda session: (session.session, session.joined_at)
        )

        schema._publications = db.attach_table(Publications)
        schema._events = db.attach_table(Events)

        return schema



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint

import numpy as np
from zlmdb import MapBytes32FlatBuffers, flatbuffers, table

from cfxdb import pack_uint256, unpack_uint256
from cfxdb.gen.xbr import Block as BlockGen


class _BlockGen(BlockGen.Block):
    """
    Expand methods on the class code generated by flatc.

    FIXME: come up with a PR for fla_CatalogGentc to generated this stuff automatically.
    """

    @classmethod
    def GetRootAsBlock(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _BlockGen()
        x.Init(buf, n + offset)
        return x

    def BlockNumberAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def BlockHashAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class Block(object):
    """
    Blockchain blocks. This table stores information about the series of Ethereum blocks that make up the blockchain.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        # uint64
        self._timestamp = None

        # [uint8] (uint256)
        self._block_number = None

        # [uint8]
        self._block_hash = None

        # uint32
        self._cnt_events = None

    def marshal(self) -> dict:
        obj = {
            "timestamp": int(self.timestamp) if self.timestamp else None,
            "block_number": pack_uint256(self.block_number) if self.block_number else 0,
            "block_hash": bytes(self.block_hash) if self.block_hash else None,
            "cnt_events": self.cnt_events,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def timestamp(self) -> np.datetime64:
        """
        Timestamp when record was inserted (Unix epoch time in ns).
        """
        if self._timestamp is None and self._from_fbs:
            self._timestamp = np.datetime64(self._from_fbs.Timestamp(), "ns")
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._timestamp = value

    @property
    def block_hash(self) -> bytes:
        """
        Block hash.
        """
        if self._block_hash is None and self._from_fbs:
            if self._from_fbs.BlockHashLength():
                self._block_hash = self._from_fbs.BlockHashAsBytes()
        return self._block_hash

    @block_hash.setter
    def block_hash(self, value: bytes):
        assert value is None or type(value) == bytes
        self._block_hash = value

    @property
    def block_number(self) -> int:
        """
        Primary key: block number.
        """
        if self._block_number is None and self._from_fbs:
            if self._from_fbs.BlockNumberLength():
                _block_number = self._from_fbs.BlockNumberAsBytes()
                self._block_number = unpack_uint256(bytes(_block_number))
            else:
                self._block_number = 0
        return self._block_number

    @block_number.setter
    def block_number(self, value: int):
        assert value is None or type(value) == int
        self._block_number = value

    @property
    def cnt_events(self) -> int:
        """
        Number of XBR blockchain log events found in the block.
        """
        if self._cnt_events is None and self._from_fbs:
            self._cnt_events = self._from_fbs.CntEvents()
        return self._cnt_events or 0

    @cnt_events.setter
    def cnt_events(self, value: int):
        assert value is None or type(value) == int
        self._cnt_events = value

    @staticmethod
    def cast(buf):
        return Block(_BlockGen.GetRootAsBlock(buf, 0))

    def build(self, builder):
        block_number = self.block_number
        if block_number:
            block_number = builder.CreateString(pack_uint256(block_number))

        block_hash = self.block_hash
        if block_hash:
            block_hash = builder.CreateString(block_hash)

        BlockGen.BlockStart(builder)

        if self.timestamp:
            BlockGen.BlockAddTimestamp(builder, int(self.timestamp))

        if block_number:
            BlockGen.BlockAddBlockNumber(builder, block_number)

        if block_hash:
            BlockGen.BlockAddBlockHash(builder, block_hash)

        if self.cnt_events:
            BlockGen.BlockAddCntEvents(builder, self.cnt_events)

        final = BlockGen.BlockEnd(builder)

        return final


@table("a4a0553e-24fa-4280-9959-5805f034d861", build=Block.build, cast=Block.cast)
class Blocks(MapBytes32FlatBuffers):
    """
    Blockchain blocks processed.

    Map :class:`zlmdb.MapBytes32FlatBuffers` from ``block_number`` to :class:`cfxdb.xbr.Block`
    """



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from cfxdb.gen.mrealm.WorkerGroupStatus import WorkerGroupStatus

STATUS_BY_CODE = {
    WorkerGroupStatus.NONE: "NONE",
    WorkerGroupStatus.STOPPED: "STOPPED",
    WorkerGroupStatus.STARTING: "STARTING",
    WorkerGroupStatus.RUNNING: "RUNNING",
    WorkerGroupStatus.PAUSED: "PAUSED",
    WorkerGroupStatus.STOPPING: "STOPPING",
    WorkerGroupStatus.ERROR: "ERROR",
    WorkerGroupStatus.DEGRADED: "DEGRADED",
}
STATUS_BY_NAME = {
    "NONE": WorkerGroupStatus.NONE,
    "STOPPED": WorkerGroupStatus.STOPPED,
    "STARTING": WorkerGroupStatus.STARTING,
    "RUNNING": WorkerGroupStatus.RUNNING,
    "PAUSED": WorkerGroupStatus.PAUSED,
    "STOPPING": WorkerGroupStatus.STOPPING,
    "ERROR": WorkerGroupStatus.ERROR,
    "DEGRADED": WorkerGroupStatus.DEGRADED,
}
STATUS_STOPPED = WorkerGroupStatus.STOPPED
STATUS_STARTING = WorkerGroupStatus.STARTING
STATUS_RUNNING = WorkerGroupStatus.RUNNING
STATUS_PAUSED = WorkerGroupStatus.PAUSED
STATUS_STOPPING = WorkerGroupStatus.STOPPING
STATUS_ERROR = WorkerGroupStatus.ERROR
STATUS_DEGRADED = WorkerGroupStatus.DEGRADED



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from cfxdb.gen.xbr.ActorType import ActorType
from cfxdb.gen.xbr.MemberLevel import MemberLevel
from cfxdb.xbr.actor import Actor, Actors
from cfxdb.xbr.api import Api, Apis
from cfxdb.xbr.block import Block, Blocks
from cfxdb.xbr.catalog import Catalog, Catalogs
from cfxdb.xbr.consent import Consent, Consents, IndexConsentByMemberAddress
from cfxdb.xbr.market import IndexMarketsByActor, IndexMarketsByMaker, IndexMarketsByOwner, Market, Markets
from cfxdb.xbr.member import Member, Members
from cfxdb.xbr.schema import Schema
from cfxdb.xbr.token import TokenApproval, TokenApprovals, TokenTransfer, TokenTransfers

__all__ = (
    # database schema
    "Schema",
    # enum types
    "MemberLevel",
    "ActorType",
    # table/index types
    "Actor",
    "Actors",
    "Api",
    "Apis",
    "Block",
    "Blocks",
    "Catalog",
    "Catalogs",
    "Consent",
    "Consents",
    "IndexConsentByMemberAddress",
    "Market",
    "Markets",
    "IndexMarketsByOwner",
    "IndexMarketsByActor",
    "IndexMarketsByMaker",
    "Member",
    "Members",
    "TokenApproval",
    "TokenApprovals",
    "TokenTransfer",
    "TokenTransfers",
)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbr

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# XBR data consent set by market actors.
class Consent(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Consent()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsConsent(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Consent
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # The ID of the catalog consent was set within [primary key].
    # Consent
    def MarketOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Consent
    def MarketOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Consent
    def MarketOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Consent
    def MarketOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Member (must be actor in market) that has set consent [primary key].
    # Consent
    def Member(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Consent
    def MemberAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Consent
    def MemberLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Consent
    def MemberIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Delegate for which consent is set [primary key].
    # Consent
    def Delegate(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Consent
    def DelegateAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Consent
    def DelegateLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Consent
    def DelegateIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # Type of delegate for which consent was granted [primary key].
    # Consent
    def DelegateType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # The ID of the catalog this consent was set for [primary key].
    # Consent
    def CatalogOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Consent
    def CatalogOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Consent
    def CatalogOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Consent
    def CatalogOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Database transaction time (epoch time in ns).
    # Consent
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Block number when the consent was updated.
    # Consent
    def Updated(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Consent
    def UpdatedAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Consent
    def UpdatedLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Consent
    def UpdatedIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        return o == 0

    # Consent granted or revoked.
    # Consent
    def Consent(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # The WAMP URI prefix to be used by the delegate when providing data service under an API of the catalog and in the data plane realm.
    # Consent
    def ServicePrefix(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Transaction hash of the transaction this change was committed to the blockchain under.
    # Consent
    def Tid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Consent
    def TidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Consent
    def TidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Consent
    def TidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        return o == 0

    # This is the signature the user (actor) supplied when setting the consent status.
    # Consent
    def Signature(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Consent
    def SignatureAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Consent
    def SignatureLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Consent
    def SignatureIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        return o == 0

    # Whether the consent was set in blockchain
    # Consent
    def Synced(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False


def ConsentStart(builder):
    builder.StartObject(12)


def Start(builder):
    ConsentStart(builder)


def ConsentAddMarketOid(builder, marketOid):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(marketOid), 0)


def AddMarketOid(builder, marketOid):
    ConsentAddMarketOid(builder, marketOid)


def ConsentStartMarketOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartMarketOidVector(builder, numElems: int) -> int:
    return ConsentStartMarketOidVector(builder, numElems)


def ConsentAddMember(builder, member):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(member), 0)


def AddMember(builder, member):
    ConsentAddMember(builder, member)


def ConsentStartMemberVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartMemberVector(builder, numElems: int) -> int:
    return ConsentStartMemberVector(builder, numElems)


def ConsentAddDelegate(builder, delegate):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(delegate), 0)


def AddDelegate(builder, delegate):
    ConsentAddDelegate(builder, delegate)


def ConsentStartDelegateVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartDelegateVector(builder, numElems: int) -> int:
    return ConsentStartDelegateVector(builder, numElems)


def ConsentAddDelegateType(builder, delegateType):
    builder.PrependUint8Slot(3, delegateType, 0)


def AddDelegateType(builder, delegateType):
    ConsentAddDelegateType(builder, delegateType)


def ConsentAddCatalogOid(builder, catalogOid):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(catalogOid), 0)


def AddCatalogOid(builder, catalogOid):
    ConsentAddCatalogOid(builder, catalogOid)


def ConsentStartCatalogOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartCatalogOidVector(builder, numElems: int) -> int:
    return ConsentStartCatalogOidVector(builder, numElems)


def ConsentAddTimestamp(builder, timestamp):
    builder.PrependUint64Slot(5, timestamp, 0)


def AddTimestamp(builder, timestamp):
    ConsentAddTimestamp(builder, timestamp)


def ConsentAddUpdated(builder, updated):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(updated), 0)


def AddUpdated(builder, updated):
    ConsentAddUpdated(builder, updated)


def ConsentStartUpdatedVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartUpdatedVector(builder, numElems: int) -> int:
    return ConsentStartUpdatedVector(builder, numElems)


def ConsentAddConsent(builder, consent):
    builder.PrependBoolSlot(7, consent, 0)


def AddConsent(builder, consent):
    ConsentAddConsent(builder, consent)


def ConsentAddServicePrefix(builder, servicePrefix):
    builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(servicePrefix), 0)


def AddServicePrefix(builder, servicePrefix):
    ConsentAddServicePrefix(builder, servicePrefix)


def ConsentAddTid(builder, tid):
    builder.PrependUOffsetTRelativeSlot(9, flatbuffers.number_types.UOffsetTFlags.py_type(tid), 0)


def AddTid(builder, tid):
    ConsentAddTid(builder, tid)


def ConsentStartTidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartTidVector(builder, numElems: int) -> int:
    return ConsentStartTidVector(builder, numElems)


def ConsentAddSignature(builder, signature):
    builder.PrependUOffsetTRelativeSlot(10, flatbuffers.number_types.UOffsetTFlags.py_type(signature), 0)


def AddSignature(builder, signature):
    ConsentAddSignature(builder, signature)


def ConsentStartSignatureVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartSignatureVector(builder, numElems: int) -> int:
    return ConsentStartSignatureVector(builder, numElems)


def ConsentAddSynced(builder, synced):
    builder.PrependBoolSlot(11, synced, 0)


def AddSynced(builder, synced):
    ConsentAddSynced(builder, synced)


def ConsentEnd(builder):
    return builder.EndObject()


def End(builder):
    return ConsentEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import struct
import uuid
from datetime import datetime
from pprint import pformat

from cfxdb.gen import oid_t
from cfxdb.gen.user import Organization as OrganizationGenFbs


class OrganizationFbs(object):
    """
    CFC user database class using Flatbuffers.
    """

    # oid: uuid.UUID

    # label: Optional[str]
    # description: Optional[str]
    # tags: Optional[List[str]]

    # name: str
    # registered: datetime
    # otype: int

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        self._oid = None

        self._label = None
        self._description = None
        self._tags = None

        self._name = None
        self._otype = None
        self._registered = None

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        if other.oid != self.oid:
            return False

        if other.label != self.label:
            return False
        if other.description != self.description:
            return False
        if (other.tags and not self.tags) or (not other.tags and self.tags):
            return False
        if other.tags:
            if not set(other.tags).intersection(set(self.tags)):
                return False

        if other.name != self.name:
            return False
        if other.otype != self.otype:
            return False
        if other.registered != self.registered:
            return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "\n{}\n".format(pformat(self.marshal()))

    def copy(self, other):
        self.oid = other.oid
        self.label = other.label
        self.description = other.description
        self.tags = other.tags
        self.name = other.name
        self.otype = other.otype
        self.registered = other.registered

        # _unknown is not copied!

    @property
    def oid(self):
        if self._oid is None and self._from_fbs:
            oid = self._from_fbs.Oid()
            if oid:
                bytes = struct.pack(">Q", oid.Lsb()) + struct.pack(">Q", oid.Msb())
                self._oid = uuid.UUID(bytes=bytes)
        return self._oid

    @oid.setter
    def oid(self, value):
        assert value is None or isinstance(value, uuid.UUID)

        self._oid = value

    @property
    def label(self):
        if self._label is None and self._from_fbs:
            s = self._from_fbs.Label()
            if s:
                self._label = s.decode()
        return self._label

    @label.setter
    def label(self, value):
        assert value is None or type(value) == str

        self._label = value

    @property
    def description(self):
        if self._description is None and self._from_fbs:
            s = self._from_fbs.Description()
            if s:
                self._description = s.decode()
        return self._description

    @description.setter
    def description(self, value):
        assert value is None or type(value) == str

        self._description = value

    @property
    def tags(self):
        if self._tags is None and self._from_fbs:
            if self._from_fbs.TagsLength() > 0:
                tags = []
                for i in range(self._from_fbs.TagsLength()):
                    s = self._from_fbs.Tags(i)
                    tags.append(s.decode())
                self._tags = tags
        return self._tags

    @tags.setter
    def tags(self, value):
        assert value is None or (type(value) == list and (type(tag) == str for tag in value))

        self._tags = value

    @property
    def name(self):
        if self._name is None and self._from_fbs:
            s = self._from_fbs.Name()
            if s:
                self._name = s.decode()
        return self._name

    @name.setter
    def name(self, value):
        assert value is None or type(value) == str

        self._name = value

    @property
    def registered(self):
        if self._registered is None and self._from_fbs:
            val = self._from_fbs.Registered()
            if val:
                # utcfromtimestamp
                # self._registered = datetime.utcfromtimestamp(float(val) / 1000000.)
                # calendar.timegm(dt.utctimetuple())
                self._registered = datetime.fromtimestamp(float(val) / 1000000.0)

        return self._registered

    @registered.setter
    def registered(self, value):
        assert value is None or isinstance(value, datetime)

        self._registered = value

    @property
    def otype(self):
        if self._otype is None and self._from_fbs:
            self._otype = self._from_fbs.Otype()

        return self._otype

    @otype.setter
    def otype(self, value):
        assert value is None or type(value) == int

        self._otype = value

    @staticmethod
    def cast(buf):
        return OrganizationFbs(OrganizationGenFbs.Organization.GetRootAsOrganization(buf, 0))

    def build(self, builder):
        # serialize all stuff we need later first (because we cannot build nested) ..

        # label: string
        label = self.label
        if label:
            label = builder.CreateString(label)

        # description: string
        description = self.description
        if description:
            description = builder.CreateString(description)

        # name: string
        name = self.name
        if name:
            name = builder.CreateString(name)

        # tags: [string]
        tags = self.tags
        if tags:
            _tags = []
            for tag in tags:
                _tags.append(builder.CreateString(tag))

            OrganizationGenFbs.OrganizationStartTagsVector(builder, len(_tags))

            for o in reversed(_tags):
                builder.PrependUOffsetTRelative(o)

            tags = builder.EndVector()

        # now start and build a new object ..
        OrganizationGenFbs.OrganizationStart(builder)

        # oid: uuid.UUID
        if self.oid:
            bytes = self.oid.bytes
            msb = struct.unpack(">Q", bytes[8:])[0]
            lsb = struct.unpack(">Q", bytes[:8])[0]
            oid = oid_t.Createoid_t(builder, msb=msb, lsb=lsb)
            OrganizationGenFbs.OrganizationAddOid(builder, oid)

        if label:
            OrganizationGenFbs.OrganizationAddLabel(builder, label)
        if description:
            OrganizationGenFbs.OrganizationAddDescription(builder, description)
        if tags:
            OrganizationGenFbs.OrganizationAddTags(builder, tags)
        if name:
            OrganizationGenFbs.OrganizationAddName(builder, name)

        # registered: datetime
        if self.registered:
            registered = int(self.registered.timestamp() * 1000000)
            OrganizationGenFbs.OrganizationAddRegistered(builder, registered)

        if self.otype:
            OrganizationGenFbs.OrganizationAddOtype(builder, self.otype)

        # finish the object.
        final = OrganizationGenFbs.OrganizationEnd(builder)

        return final



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import hashlib
import os

import multihash


def _gen_ipfs_hash():
    # IPFS uses sha2-256 and B58 string encoding, eg "QmevbQAb6L7MsZUxqsspYREqMhMqaaV6arvF4UytgdsSfX"
    data = os.urandom(100)
    digest = hashlib.sha256(data).digest()
    hash = multihash.encode(digest, "sha2-256")
    return multihash.to_b58_string(hash)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: meta

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class Comment(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Comment()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsComment(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Comment
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Comment OID.
    # Comment
    def Oid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Comment
    def OidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Comment
    def OidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Comment
    def OidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Table of the object commented on.
    # Comment
    def TableOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Comment
    def TableOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Comment
    def TableOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Comment
    def TableOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Object (within the table) commented on.
    # Comment
    def ObjectOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Comment
    def ObjectOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Comment
    def ObjectOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Comment
    def ObjectOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # User that authored this comment.
    # Comment
    def AuthorOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Comment
    def AuthorOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Comment
    def AuthorOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Comment
    def AuthorOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # When this comment is a reply, the OID of the comment replied to.
    # Comment
    def ReplyTo(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Comment
    def ReplyToAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Comment
    def ReplyToLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Comment
    def ReplyToIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # The text of the comment.
    # Comment
    def Comment(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Timestamp when the comment was posted.
    # Comment
    def Posted(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0


def CommentStart(builder):
    builder.StartObject(7)


def Start(builder):
    CommentStart(builder)


def CommentAddOid(builder, oid):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(oid), 0)


def AddOid(builder, oid):
    CommentAddOid(builder, oid)


def CommentStartOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartOidVector(builder, numElems: int) -> int:
    return CommentStartOidVector(builder, numElems)


def CommentAddTableOid(builder, tableOid):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(tableOid), 0)


def AddTableOid(builder, tableOid):
    CommentAddTableOid(builder, tableOid)


def CommentStartTableOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartTableOidVector(builder, numElems: int) -> int:
    return CommentStartTableOidVector(builder, numElems)


def CommentAddObjectOid(builder, objectOid):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(objectOid), 0)


def AddObjectOid(builder, objectOid):
    CommentAddObjectOid(builder, objectOid)


def CommentStartObjectOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartObjectOidVector(builder, numElems: int) -> int:
    return CommentStartObjectOidVector(builder, numElems)


def CommentAddAuthorOid(builder, authorOid):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(authorOid), 0)


def AddAuthorOid(builder, authorOid):
    CommentAddAuthorOid(builder, authorOid)


def CommentStartAuthorOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartAuthorOidVector(builder, numElems: int) -> int:
    return CommentStartAuthorOidVector(builder, numElems)


def CommentAddReplyTo(builder, replyTo):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(replyTo), 0)


def AddReplyTo(builder, replyTo):
    CommentAddReplyTo(builder, replyTo)


def CommentStartReplyToVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartReplyToVector(builder, numElems: int) -> int:
    return CommentStartReplyToVector(builder, numElems)


def CommentAddComment(builder, comment):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(comment), 0)


def AddComment(builder, comment):
    CommentAddComment(builder, comment)


def CommentAddPosted(builder, posted):
    builder.PrependUint64Slot(6, posted, 0)


def AddPosted(builder, posted):
    CommentAddPosted(builder, posted)


def CommentEnd(builder):
    return builder.EndObject()


def End(builder):
    return CommentEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from cfxdb.xbr.consent import Consents, IndexConsentByMemberAddress
from cfxdb.xbrmm.channel import (
    IndexPayingChannelByDelegate,
    IndexPayingChannelByRecipient,
    IndexPaymentChannelByActor,
    IndexPaymentChannelByDelegate,
    PayingChannelBalances,
    PayingChannels,
    PaymentChannelBalances,
    PaymentChannels,
)
from cfxdb.xbrmm.ipfs_file import IPFSFiles
from cfxdb.xbrmm.offer import IndexOfferByKey, Offers
from cfxdb.xbrmm.transaction import Transactions
from cfxdb.xbrmm.userkey import IndexUserKeyByMember, UserKeys


class Schema(object):
    """
    CFC edge database schema for ZLMDB.
    """

    def __init__(self, db):
        self.db = db

    consents: Consents
    """
    XBR data consents.
    """

    idx_consent_by_member_adr: IndexConsentByMemberAddress
    """
    Consents-by-members-address index with ``(member_adr|bytes[20], joined|int) -> member_adr|UUID`` mapping.
    """

    payment_channels: PaymentChannels
    """
    Payment channels for XBR consumer delegates.
    """

    idx_payment_channel_by_delegate: IndexPaymentChannelByDelegate
    """
    Maps from XBR consumer delegate address to the currently active payment
    channel address for the given consumer delegate.
    """

    idx_payment_channel_by_actor: IndexPaymentChannelByActor
    """
    Maps from XBR consumer actor address to the currently active payment
    channel address for the given consumer actor.
    """

    payment_balances: PaymentChannelBalances
    """
    Current off-chain balances within payment channels.
    """

    paying_channels: PayingChannels
    """
    Paying channels for XBR provider delegates.
    """

    idx_paying_channel_by_delegate: IndexPayingChannelByDelegate
    """
    Maps from XBR provider delegate address to the currently active paying
    channel address for the given provider delegate.
    """

    idx_paying_channel_by_recipient: IndexPayingChannelByRecipient
    """
    Maps from XBR recipient address to the currently active paying
    channel address for the given recipient.
    """

    paying_balances: PayingChannelBalances
    """
    Current off-chain balances within paying channels.
    """

    offers: Offers
    """
    Data encryption key offers.
    """

    idx_offer_by_key: IndexOfferByKey
    """
    Index of key offers by key ID (rather than offer ID, as the object table
    is indexed by).
    """

    transactions: Transactions
    """
    """

    user_keys: UserKeys
    """
    User client keys database table :class:`xbrmm.UserKeys`.
    """

    idx_user_key_by_member: IndexUserKeyByMember
    """
    Index "by pubkey" of user keys :class:`xbrmm.IndexUserKeyByMember`.
    """

    ipfs_files: IPFSFiles
    """
    IPFS files download log table :class:`xbrmm.IPFSFiles`.
    """

    @staticmethod
    def attach(db):
        """
        Factory to create a schema from attaching to a database. The schema tables
        will be automatically mapped as persistant maps and attached to the
        database slots.

        :param db: zlmdb.Database
        :return: object of Schema
        """
        schema = Schema(db)

        schema.consents = db.attach_table(Consents)

        schema.idx_consent_by_member_adr = db.attach_table(IndexConsentByMemberAddress)

        schema.payment_channels = db.attach_table(PaymentChannels)

        schema.idx_payment_channel_by_delegate = db.attach_table(IndexPaymentChannelByDelegate)
        schema.payment_channels.attach_index(
            "idx1",
            schema.idx_payment_channel_by_delegate,
            lambda channel: (bytes(channel.delegate), channel.timestamp),
        )

        schema.idx_payment_channel_by_actor = db.attach_table(IndexPaymentChannelByActor)
        schema.payment_channels.attach_index(
            "idx2", schema.idx_payment_channel_by_actor, lambda channel: (bytes(channel.actor), channel.timestamp)
        )

        schema.payment_balances = db.attach_table(PaymentChannelBalances)

        schema.paying_channels = db.attach_table(PayingChannels)

        schema.idx_paying_channel_by_delegate = db.attach_table(IndexPayingChannelByDelegate)
        schema.paying_channels.attach_index(
            "idx1", schema.idx_paying_channel_by_delegate, lambda channel: (bytes(channel.delegate), channel.timestamp)
        )

        schema.idx_paying_channel_by_recipient = db.attach_table(IndexPayingChannelByRecipient)
        schema.paying_channels.attach_index(
            "idx2",
            schema.idx_paying_channel_by_recipient,
            lambda channel: (bytes(channel.recipient), channel.timestamp),
        )

        schema.paying_balances = db.attach_table(PayingChannelBalances)

        schema.offers = db.attach_table(Offers)
        schema.idx_offer_by_key = db.attach_table(IndexOfferByKey)
        schema.offers.attach_index("idx1", schema.idx_offer_by_key, lambda offer: offer.key)

        schema.user_keys = db.attach_table(UserKeys)

        schema.idx_user_key_by_member = db.attach_table(IndexUserKeyByMember)
        schema.user_keys.attach_index(
            "idx1", schema.idx_user_key_by_member, lambda user_key: (user_key.owner, user_key.created)
        )

        schema.transactions = db.attach_table(Transactions)

        schema.ipfs_files = db.attach_table(IPFSFiles)

        return schema



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from zlmdb import MapTimestampUuidStringFlatBuffers, table

from cfxdb.log.mworker_log import MWorkerLog


@table("5ceaa500-4832-451c-adf4-4fc4968cece0", build=MWorkerLog.build, cast=MWorkerLog.cast)
class MWorkerLogs(MapTimestampUuidStringFlatBuffers):
    """
    Managed node worker heartbeat log records (``(timestamp, node_id, worker_id) -> worker_log``).
    """



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: mrealm

# Status of RouterWorkerGroups.
class WorkerGroupStatus(object):
    # Unset status.
    NONE = 0
    # Initial status of a newly created worker group.
    STOPPED = 1
    # The worker group is currently starting.
    STARTING = 2
    # The worker group is currently running.
    RUNNING = 3
    # The worker group is currently paused (it is serving, but the watch/change engine doesn't operate).
    PAUSED = 4
    # The worker group is currently stopping.
    STOPPING = 5
    # The worker group is stopped in an error state.
    ERROR = 6
    # The worker group is currently running, but degraded (not all components and services desired are actually running at this moment, but some have failed, but may be retried automatically).
    DEGRADED = 7



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import json
import os
import sys
import uuid
from pprint import pprint
from typing import Any, Dict, List

import cbor2
import click
import numpy as np
import zlmdb
from autobahn.wamp.serializer import JsonObjectSerializer
from pygments import formatters, highlight, lexers
from txaio import time_ns

import cfxdb
from cfxdb.xbrnetwork import Account, UserKey


def pprint_json(data):
    json_str = json.dumps(data, separators=(", ", ": "), sort_keys=True, indent=4, ensure_ascii=False)
    console_str = highlight(json_str, lexers.JsonLexer(), formatters.Terminal256Formatter(style="fruity"))
    print(console_str)


class Exporter(object):
    """
    CFXDB database exporter.
    """

    def __init__(self, dbpath):
        """

        :param dbpath: Database file to open.
        """
        self._dbpath = os.path.abspath(dbpath)
        self._db = zlmdb.Database(dbpath=self._dbpath, maxsize=2**30, readonly=False)
        self._db.__enter__()

        self._schemata: Dict[str, Any] = {}
        self._schema_tables = {}

        if False:
            self._meta = cfxdb.meta.Schema.attach(self._db)
            self._globalschema = cfxdb.globalschema.GlobalSchema.attach(self._db)
            self._mrealmschema = cfxdb.mrealmschema.MrealmSchema.attach(self._db)
            self._xbr = cfxdb.xbr.Schema.attach(self._db)
            self._xbrmm = cfxdb.xbrmm.Schema.attach(self._db)
            self._xbrnetwork = cfxdb.xbrnetwork.Schema.attach(self._db)

            self._schemata = {
                "meta": self._meta,
                "globalschema": self._globalschema,
                "mrealmschema": self._mrealmschema,
                "xbr": self._xbr,
                "xbrmm": self._xbrmm,
                "xbrnetwork": self._xbrnetwork,
            }
            self._schema_tables = {}
            for schema_name, schema in self._schemata.items():
                tables = {}
                first = None
                for k, v in schema.__annotations__.items():
                    for line in v.__doc__.splitlines():
                        line = line.strip()
                        if line != "":
                            first = line[:80]
                            break
                    tables[k] = first
                self._schema_tables[schema_name] = tables

    @property
    def dbpath(self) -> str:
        """

        :return:
        """
        return self._dbpath

    def schemata(self) -> List[str]:
        """

        :return:
        """
        return sorted(self._schemata.keys())

    def tables(self, schema_name):
        """

        :param schema_name:
        :return:
        """
        if schema_name in self._schema_tables:
            return sorted(self._schema_tables[schema_name].keys())
        else:
            return None

    def table_docs(self, schema_name, table_name):
        """

        :param schema_name:
        :param table_name:
        :return:
        """
        if schema_name in self._schema_tables and table_name in self._schema_tables[schema_name]:
            return self._schema_tables[schema_name][table_name]
        else:
            return None

    def _add_test_data(self):
        account = Account()
        account.oid = uuid.uuid4()
        account.created = np.datetime64(time_ns(), "ns")
        account.username = "alice"
        account.email = "alice@example.com"
        account.wallet_type = 2  # metamask
        account.wallet_address = os.urandom(20)
        # account.wallet_address = binascii.a2b_hex('f5173a6111B2A6B3C20fceD53B2A8405EC142bF6')

        userkey = UserKey()
        userkey.pubkey = os.urandom(32)
        # userkey.pubkey = binascii.a2b_hex('b7e6462121b9632b2bfcc5a3beef0b49dd865093ad003d011d4abbb68476d5b4')
        userkey.created = account.created
        userkey.owner = account.oid

        with self._db.begin(write=True) as txn:
            self._xbrnetwork.accounts[txn, account.oid] = account
            self._xbrnetwork.user_keys[txn, userkey.pubkey] = userkey

    def print_slots(self, include_description=False):
        """

        :param include_description:
        :return:
        """
        print(click.style("Database slots:\n", fg="white", bold=True))
        slots = self._db._get_slots()
        for slot_id in slots:
            slot = slots[slot_id]

            with self._db.begin() as txn:
                pmap = zlmdb.PersistentMap(slot.slot)
                records = pmap.count(txn)

            if include_description:
                print(
                    "   Table in slot {} ({}) with {} records is bound to class {}: {}".format(
                        click.style(str(slot.slot), fg="yellow", bold=True),
                        click.style(str(slot_id), fg="white"),
                        click.style(str(records), fg="yellow", bold=True),
                        click.style(slot.name, fg="yellow"),
                        slot.description,
                    )
                )
            else:
                print(
                    "   Table in slot {} ({}) with {} records is bound to class {}".format(
                        click.style(str(slot.slot), fg="yellow", bold=True),
                        click.style(str(slot_id), fg="white"),
                        click.style(str(records), fg="yellow", bold=True),
                        click.style(slot.name, fg="yellow"),
                    )
                )
        print("")

    def print_stats(self, include_slots=False):
        """

        :return:
        """
        print(click.style("Database statistics:\n", fg="white", bold=True))
        pprint_json(self._db.stats(include_slots=include_slots))
        print("")

    def print_config(self):
        print(click.style("Database configuration:\n", fg="white", bold=True))
        pprint_json(self._db.config())
        print("")

    def export_database(
        self,
        filename=None,
        include_indexes=False,
        include_schemata=None,
        exclude_tables=None,
        use_json=False,
        quiet=False,
        use_binary_hex_encoding=False,
    ):
        """

        :param filename:
        :param include_indexes:
        :param include_schemata:
        :param exclude_tables:
        :param use_json:
        :param use_binary_hex_encoding:
        :returns:
        """
        if include_schemata is None:
            schemata = sorted(self._schemata.keys())
        else:
            assert type(include_schemata) == list
            schemata = sorted(list(set(include_schemata).intersection(self._schemata.keys())))

        if exclude_tables is None:
            exclude_tables = set()
        else:
            assert type(exclude_tables) == list
            exclude_tables = set(exclude_tables)

        result = {}
        with self._db.begin() as txn:
            for schema_name in schemata:
                for table_name in self._schema_tables[schema_name]:
                    fq_table_name = "{}.{}".format(schema_name, table_name)
                    if fq_table_name not in exclude_tables:
                        table = self._schemata[schema_name].__dict__[table_name]
                        if not table.is_index() or include_indexes:
                            recs = []
                            for key, val in table.select(txn, return_keys=True, return_values=True):
                                if val:
                                    if hasattr(val, "marshal"):
                                        val = val.marshal()
                                recs.append((table._serialize_key(key), val))
                            if recs:
                                if schema_name not in result:
                                    result[schema_name] = {}
                                result[schema_name][table_name] = recs

        if use_json:
            ser = JsonObjectSerializer(batched=False, use_binary_hex_encoding=use_binary_hex_encoding)
            try:
                data: bytes = ser.serialize(result)
            except TypeError as e:
                print(e)
                pprint(result)
                sys.exit(1)
        else:
            data: bytes = cbor2.dumps(result)

        if filename:
            with open(filename, "wb") as f:
                f.write(data)
        else:
            sys.stdout.buffer.write(data)

        if not quiet:
            print(
                '\nExported database [dbpath="{dbpath}", filename="{filename}", filesize={filesize}]:\n'.format(
                    dbpath=click.style(self._dbpath, fg="yellow"),
                    filename=click.style(filename, fg="yellow"),
                    filesize=click.style(len(data), fg="yellow"),
                )
            )
            for schema_name in result:
                for table_name in result[schema_name]:
                    cnt = len(result[schema_name][table_name])
                    if cnt:
                        print(
                            "{:.<52}: {}".format(
                                click.style("{}.{}".format(schema_name, table_name), fg="white", bold=True),
                                click.style(str(cnt) + " records", fg="yellow"),
                            )
                        )

    def import_database(
        self,
        filename=None,
        include_indexes=False,
        include_schemata=None,
        exclude_tables=None,
        use_json=False,
        quiet=False,
        use_binary_hex_encoding=False,
    ):
        """

        :param filename:
        :param include_indexes:
        :param include_schemata:
        :param exclude_tables:
        :param use_json:
        :param use_binary_hex_encoding:
        :returns:
        """
        if include_schemata is None:
            schemata = sorted(self._schemata.keys())
        else:
            assert type(include_schemata) == list
            schemata = sorted(list(set(include_schemata).intersection(self._schemata.keys())))

        if exclude_tables is None:
            exclude_tables = set()
        else:
            assert type(exclude_tables) == list
            exclude_tables = set(exclude_tables)

        if filename:
            with open(filename, "rb") as f:
                data = f.read()
        else:
            data = sys.stdin.read()

        if use_json:
            ser = JsonObjectSerializer(batched=False, use_binary_hex_encoding=use_binary_hex_encoding)
            db_data = ser.unserialize(data)[0]
        else:
            db_data = cbor2.loads(data)

        if not quiet:
            print(
                '\nImporting database [dbpath="{dbpath}", filename="{filename}", filesize={filesize}]:\n'.format(
                    dbpath=self._dbpath, filename=filename, filesize=len(data)
                )
            )

        with self._db.begin(write=True) as txn:
            for schema_name in schemata:
                for table_name in self._schema_tables[schema_name]:
                    fq_table_name = "{}.{}".format(schema_name, table_name)
                    if fq_table_name not in exclude_tables:
                        table = self._schemata[schema_name].__dict__[table_name]
                        if not table.is_index() or include_indexes:
                            if schema_name in db_data and table_name in db_data[schema_name]:
                                cnt = 0
                                for key, val in db_data[schema_name][table_name]:
                                    key = table._deserialize_key(key)
                                    val = table.parse(val)
                                    table[txn, key] = val
                                    cnt += 1
                                if cnt and not quiet:
                                    print(
                                        "{:.<52}: {}".format(
                                            click.style(
                                                "{}.{}".format(schema_name, table_name), fg="white", bold=True
                                            ),
                                            click.style(str(cnt) + " records", fg="yellow"),
                                        )
                                    )
                            else:
                                if not quiet:
                                    print("No data to import for {}.{}!".format(schema_name, table_name))



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbrmm

# Type of a XBR off-chain channel: paying channel (for provider delegates selling data services) or payment channel (for consumer delegates buying data services).
class ChannelType(object):
    # Unset
    NONE = 0
    # Payment channel: from buyer/consumer delegate to maker maker.
    PAYMENT = 1
    # Paying channel: from market maker to seller/provider delegate.
    PAYING = 2



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint

from zlmdb import MapBytes32FlatBuffers, flatbuffers, table

from cfxdb import pack_uint256, unpack_uint256
from cfxdb.gen.xbr import TokenApproval as TokenApprovalGen
from cfxdb.gen.xbr import TokenTransfer as TokenTransferGen


class _TokenApprovalGen(TokenApprovalGen.TokenApproval):
    """
    Expand methods on the class code generated by flatc.

    FIXME: come up with a PR for flatc to generated this stuff automatically.
    """

    @classmethod
    def GetRootAsTokenApproval(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _TokenApprovalGen()
        x.Init(buf, n + offset)
        return x

    def TxHashAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def BlockHashAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def OwnerAddressAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def SpenderAddressAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def ValueAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class TokenApproval(object):
    """
    ``IERC20.TokenApproval`` event database object.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        self._tx_hash = None
        self._block_hash = None
        self._owner_address = None
        self._spender_address = None
        self._value = None

    def marshal(self) -> dict:
        obj = {
            "tx_hash": bytes(self._tx_hash) if self._tx_hash else None,
            "block_hash": bytes(self._block_hash) if self._block_hash else None,
            "from_address": bytes(self._owner_address) if self._owner_address else None,
            "to_address": bytes(self._spender_address) if self._spender_address else None,
            "value": self._value,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def tx_hash(self) -> bytes:
        """
        Primary key: Transaction hash.
        """
        if self._tx_hash is None and self._from_fbs:
            if self._from_fbs.TxHashLength():
                self._tx_hash = self._from_fbs.TxHashAsBytes()
        return self._tx_hash

    @tx_hash.setter
    def tx_hash(self, value: bytes):
        assert value is None or type(value) == bytes
        self._tx_hash = value

    @property
    def block_hash(self) -> bytes:
        """
        Block hash.
        """
        if self._block_hash is None and self._from_fbs:
            if self._from_fbs.TxHashLength():
                self._block_hash = self._from_fbs.BlockHashAsBytes()
        return self._block_hash

    @block_hash.setter
    def block_hash(self, value: bytes):
        assert value is None or type(value) == bytes
        self._block_hash = value

    @property
    def owner_address(self) -> bytes:
        """
        XBR token sending address.
        """
        if self._owner_address is None and self._from_fbs:
            if self._from_fbs.OwnerAddressLength():
                self._owner_address = self._from_fbs.OwnerAddressAsBytes()
        return self._owner_address

    @owner_address.setter
    def owner_address(self, value: bytes):
        assert value is None or type(value) == bytes
        self._owner_address = value

    @property
    def spender_address(self) -> bytes:
        """
        XBR token receiving address.
        """
        if self._spender_address is None and self._from_fbs:
            if self._from_fbs.TxHashLength():
                self._spender_address = self._from_fbs.SpenderAddressAsBytes()
        return self._spender_address

    @spender_address.setter
    def spender_address(self, value: bytes):
        assert value is None or type(value) == bytes
        self._spender_address = value

    @property
    def value(self) -> int:
        """
        XBR token transferred.
        """
        if self._value is None and self._from_fbs:
            if self._from_fbs.ValueLength():
                _value = self._from_fbs.ValueAsBytes()
                self._value = unpack_uint256(bytes(_value))
            else:
                self._value = 0
        return self._value

    @value.setter
    def value(self, value: int):
        assert value is None or type(value) == int
        self._value = value

    @staticmethod
    def cast(buf):
        return TokenApproval(_TokenApprovalGen.GetRootAsTokenApproval(buf, 0))

    def build(self, builder):
        tx_hash = self.tx_hash
        if tx_hash:
            tx_hash = builder.CreateString(tx_hash)

        block_hash = self.block_hash
        if block_hash:
            block_hash = builder.CreateString(block_hash)

        owner_address = self.owner_address
        if owner_address:
            owner_address = builder.CreateString(owner_address)

        spender_address = self.spender_address
        if spender_address:
            spender_address = builder.CreateString(spender_address)

        value = self.value
        if value:
            value = builder.CreateString(pack_uint256(value))

        TokenApprovalGen.TokenApprovalStart(builder)

        if tx_hash:
            TokenApprovalGen.TokenApprovalAddTxHash(builder, tx_hash)

        if block_hash:
            TokenApprovalGen.TokenApprovalAddBlockHash(builder, block_hash)

        if owner_address:
            TokenApprovalGen.TokenApprovalAddOwnerAddress(builder, owner_address)

        if spender_address:
            TokenApprovalGen.TokenApprovalAddSpenderAddress(builder, spender_address)

        if value:
            TokenApprovalGen.TokenApprovalAddValue(builder, value)

        final = TokenApprovalGen.TokenApprovalEnd(builder)

        return final


@table("f3d1f052-7160-48a7-b36a-221035fc9373", build=TokenApproval.build, cast=TokenApproval.cast)
class TokenApprovals(MapBytes32FlatBuffers):
    """
    Persisted token approvals archive.

    Map :class:`zlmdb.MapBytes32FlatBuffers` from ``transaction_hash`` to :class:`cfxdb.xbr.TokenApproval`
    """


class _TokenTransferGen(TokenTransferGen.TokenTransfer):
    """
    Expand methods on the class code generated by flatc.

    FIXME: come up with a PR for flatc to generated this stuff automatically.
    """

    @classmethod
    def GetRootAsTokenTransfer(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _TokenTransferGen()
        x.Init(buf, n + offset)
        return x

    def TxHashAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def BlockHashAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def FromAddressAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def ToAddressAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def ValueAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class TokenTransfer(object):
    """
    ``IERC20.TokenTransfer`` event database object.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        self._tx_hash = None
        self._block_hash = None
        self._from_address = None
        self._to_address = None
        self._value = None

    def marshal(self) -> dict:
        obj = {
            "tx_hash": self._tx_hash,
            "block_hash": self._block_hash,
            "from_address": self._from_address,
            "to_address": self._to_address,
            "value": self._value,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def tx_hash(self) -> bytes:
        """
        Primary key: Transaction hash.
        """
        if self._tx_hash is None and self._from_fbs:
            if self._from_fbs.TxHashLength():
                self._tx_hash = self._from_fbs.TxHashAsBytes()
        return self._tx_hash

    @tx_hash.setter
    def tx_hash(self, value: bytes):
        assert value is None or type(value) == bytes
        self._tx_hash = value

    @property
    def block_hash(self) -> bytes:
        """
        Block hash.
        """
        if self._block_hash is None and self._from_fbs:
            if self._from_fbs.TxHashLength():
                self._block_hash = self._from_fbs.BlockHashAsBytes()
        return self._block_hash

    @block_hash.setter
    def block_hash(self, value):
        assert value is None or type(value) == bytes
        self._block_hash = value

    @property
    def from_address(self) -> bytes:
        """
        XBR token sending address.
        """
        if self._from_address is None and self._from_fbs:
            if self._from_fbs.TxHashLength():
                self._from_address = self._from_fbs.FromAddressAsBytes()
        return self._from_address

    @from_address.setter
    def from_address(self, value: bytes):
        assert value is None or type(value) == bytes
        self._from_address = value

    @property
    def to_address(self) -> bytes:
        """
        XBR token receiving address.
        """
        if self._to_address is None and self._from_fbs:
            if self._from_fbs.TxHashLength():
                self._to_address = self._from_fbs.ToAddressAsBytes()
        return self._to_address

    @to_address.setter
    def to_address(self, value: bytes):
        assert value is None or type(value) == bytes
        self._to_address = value

    @property
    def value(self) -> int:
        """
        XBR token transferred.
        """
        if self._value is None and self._from_fbs:
            if self._from_fbs.TxHashLength():
                _value = self._from_fbs.ValueAsBytes()
                self._value = unpack_uint256(bytes(_value))
            else:
                self._value = 0
        return self._value

    @value.setter
    def value(self, value: int):
        assert value is None or type(value) == int
        self._value = value

    @staticmethod
    def cast(buf):
        return TokenTransfer(_TokenTransferGen.GetRootAsTokenTransfer(buf, 0))

    def build(self, builder):
        tx_hash = self.tx_hash
        if tx_hash:
            tx_hash = builder.CreateString(tx_hash)

        block_hash = self.block_hash
        if block_hash:
            block_hash = builder.CreateString(block_hash)

        from_address = self.from_address
        if from_address:
            from_address = builder.CreateString(from_address)

        to_address = self.to_address
        if to_address:
            to_address = builder.CreateString(to_address)

        value = self.value
        if value:
            value = builder.CreateString(pack_uint256(value))

        TokenTransferGen.TokenTransferStart(builder)

        if tx_hash:
            TokenTransferGen.TokenTransferAddTxHash(builder, tx_hash)

        if block_hash:
            TokenTransferGen.TokenTransferAddBlockHash(builder, block_hash)

        if from_address:
            TokenTransferGen.TokenTransferAddFromAddress(builder, from_address)

        if to_address:
            TokenTransferGen.TokenTransferAddToAddress(builder, to_address)

        if value:
            TokenTransferGen.TokenTransferAddValue(builder, value)

        final = TokenTransferGen.TokenTransferEnd(builder)

        return final


@table("7dcfffd2-4859-45d6-b9cd-ba7930a7a107", build=TokenTransfer.build, cast=TokenTransfer.cast)
class TokenTransfers(MapBytes32FlatBuffers):
    """
    Persisted token transfers archive.

    Map :class:`zlmdb.MapBytes32FlatBuffers` from ``transaction_hash`` to :class:`cfxdb.xbr.TokenTransfer`
    """



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import binascii
import os
import uuid
from datetime import datetime

import cbor2
import pytest
import txaio
from zlmdb import flatbuffers

from cfxdb.user import User, UserFbs

txaio.use_twisted()


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


def fill_user(user):
    user.oid = uuid.uuid4()

    user.label = "Homer the 3rd"
    user.description = "My motto as a user is: never read the f** manual;)"
    user.tags = ["geek", "pythonista", "lemon"]

    user.email = "homer.simpson@example.com"
    user.registered = datetime.utcnow()
    user.pubkey = binascii.b2a_hex(os.urandom(32)).decode()


@pytest.fixture(scope="function")
def user_fbs():
    _user = UserFbs()
    fill_user(_user)
    return _user


@pytest.fixture(scope="function")
def user_cbor():
    _user = User()
    fill_user(_user)
    return _user


def test_user_fbs_roundtrip(user_fbs, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = user_fbs.build(builder)
    builder.Finish(obj)
    data = builder.Output()

    # create python object from bytes (flatbuffes)
    _user = UserFbs.cast(data)

    # assert _user == user_fbs

    assert _user.oid == user_fbs.oid
    assert _user.label == user_fbs.label
    assert _user.description == user_fbs.description
    assert _user.tags == user_fbs.tags
    assert _user.email == user_fbs.email
    assert _user.registered == user_fbs.registered
    assert _user.pubkey == user_fbs.pubkey


def test_user_cbor_roundtrip(user_cbor):
    # serialize to bytes (cbor) from python object
    obj = user_cbor.marshal()
    data = cbor2.dumps(obj)

    # create python object from bytes (cbor)
    _obj = cbor2.loads(data)
    _user = User.parse(_obj)

    # assert _user == user_cbor

    assert _user.oid == user_cbor.oid
    assert _user.label == user_cbor.label
    assert _user.description == user_cbor.description
    assert _user.tags == user_cbor.tags
    assert _user.email == user_cbor.email
    assert _user.registered == user_cbor.registered
    assert _user.pubkey == user_cbor.pubkey


def test_user_copy_cbor_to_fbs(user_cbor):
    user_fbs = UserFbs()
    user_fbs.copy(user_cbor)

    # assert user_fbs == user_cbor

    assert user_fbs.oid == user_cbor.oid
    assert user_fbs.label == user_cbor.label
    assert user_fbs.description == user_cbor.description
    assert user_fbs.tags == user_cbor.tags
    assert user_fbs.email == user_cbor.email
    assert user_fbs.registered == user_cbor.registered
    assert user_fbs.pubkey == user_cbor.pubkey


def test_user_copy_fbs_to_cbor(user_fbs):
    user_cbor = User()
    user_cbor.copy(user_fbs)

    # assert user_fbs == user_cbor

    assert user_fbs.oid == user_cbor.oid
    assert user_fbs.label == user_cbor.label
    assert user_fbs.description == user_cbor.description
    assert user_fbs.tags == user_cbor.tags
    assert user_fbs.email == user_cbor.email
    assert user_fbs.registered == user_cbor.registered
    assert user_fbs.pubkey == user_cbor.pubkey



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: user

# User activation types (built into code, unconfigurable)
class ActivationType(object):
    NONE = 0
    REGISTRATION = 1
    LOGIN = 2



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: arealm

# WAMP URI check level applied in an application realm.
class UriCheckLevel(object):
    # No check level.
    NONE = 0
    # Check level "strict".
    STRICT = 1
    # Check level "loose".
    LOOSE = 2



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from pprint import pformat

from cfxdb.gen.user.UserRole import UserRole


class UserMrealmRole(object):
    """
    Database class for CFC user roles on a management realm using CBOR.
    """

    def __init__(self, roles=None, _unknown=None):
        self.roles = roles

        # private member with unknown/untouched data passing through
        self._unknown = _unknown

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if other.roles != self.roles:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "\n{}\n".format(pformat(self.marshal()))

    def copy(self, other):
        self.roles = other.roles[:]

        # _unknown is not copied!

    def marshal(self):
        assert type(self.roles) == list
        for role in self.roles:
            assert type(role) == int
            assert role in [UserRole.OWNER, UserRole.ADMIN, UserRole.USER, UserRole.GUEST]

        obj = {
            "roles": self.roles,
        }

        if self._unknown:
            # pass through all attributes unknown
            obj.update(self._unknown)
        else:
            return obj

        return obj

    @staticmethod
    def parse(data):
        assert type(data) == dict

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = {}
        for k in data:
            if k not in ["roles"]:
                _unknown[k] = data[k]

        roles = data.get("roles", None)
        assert type(roles) == list

        for role in roles:
            assert type(role) == int
            assert role in [UserRole.OWNER, UserRole.ADMIN, UserRole.USER, UserRole.GUEST]

        obj = UserMrealmRole(roles=roles, _unknown=_unknown)
        return obj



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbrnetwork

# XBR Account level.
class AccountLevel(object):
    # Unset
    NONE = 0
    # Account is active.
    ACTIVE = 1
    # Account is active and verified.
    VERIFIED = 2
    # Account is retired.
    RETIRED = 3
    # Account is subject to a temporary penalty.
    PENALTY = 4
    # Account is currently blocked and cannot current actively participate in the market.
    BLOCKED = 5



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import zlmdb

from cfxdb.cookiestore._cookie import Cookies, IndexCookiesByValue


class CookieStoreSchema(object):
    """
    Persistent cookie store.
    """

    def __init__(self, db):
        self.db = db

    cookies: Cookies
    """
    Cookies persisted in this cookie store.
    """

    idx_cookies_by_value: IndexCookiesByValue
    """
    Index of cookies by cookie value.
    """

    @staticmethod
    def attach(db: zlmdb.Database) -> "CookieStoreSchema":
        """
        Factory to create a schema from attaching to a database. The schema tables
        will be automatically mapped as persistent maps and attached to the
        database slots.
        """
        schema = CookieStoreSchema(db)

        schema.cookies = db.attach_table(Cookies)

        schema.idx_cookies_by_value = db.attach_table(IndexCookiesByValue)
        schema.cookies.attach_index("idx1", schema.idx_cookies_by_value, lambda cookie: cookie.value)

        return schema



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from datetime import datetime
from pprint import pformat
from typing import List, Optional
from uuid import UUID

from cfxdb.common import ConfigurationElement


class User(ConfigurationElement):
    """
    Users registered with this master instance.

    .. note::
        The user database exists "globally" (master-wide) and
        independent of management realms. A given user can be owner or authorized to
        access different management realms or resources therein.
    """

    def __init__(
        self,
        oid: Optional[UUID] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        email: Optional[str] = None,
        registered: Optional[datetime] = None,
        pubkey: Optional[str] = None,
        _unknown=None,
    ):
        """

        :param oid: Object ID of the user

        :param label: Optional user label of the user

        :param description: Optional user description of the user

        :param tags: Optional list of user tags on the user

        :param email: User email

        :param registered: Timestamp when the user registered

        :param pubkey: Public key of user (HEX encoded Ed25519 32 byte public key).
        """

        ConfigurationElement.__init__(self, oid=oid, label=label, description=description, tags=tags)

        self.email = email
        self.registered = registered
        self.pubkey = pubkey

        # private member with unknown/untouched data passing through
        self._unknown = _unknown

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        if not ConfigurationElement.__eq__(self, other):
            return False

        if other.email != self.email:
            return False
        if other.registered != self.registered:
            return False
        if other.pubkey != self.pubkey:
            return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "\n{}\n".format(pformat(self.marshal()))

    def copy(self, other):
        self.oid = other.oid
        self.label = other.label
        self.description = other.description
        self.tags = other.tags
        self.email = other.email
        self.registered = other.registered
        self.pubkey = other.pubkey

        # _unknown is not copied!

    def marshal(self):
        obj = ConfigurationElement.marshal(self)

        assert self.email is None or type(self.email) == str
        assert self.registered is None or isinstance(self.registered, datetime)
        assert self.pubkey is None or (type(self.pubkey) == str and len(self.pubkey) == 64)

        registered = int(self.registered.timestamp() * 1000000) if self.registered else None
        obj.update(
            {
                "email": self.email,
                "registered": registered,
                "pubkey": self.pubkey,
            }
        )

        if self._unknown:
            # pass through all attributes unknown
            obj.update(self._unknown)

        return obj

    @staticmethod
    def parse(data):
        assert type(data) == dict

        obj = ConfigurationElement.parse(data)
        data = obj._unknown

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = {}
        for k in data:
            if k not in ["email", "registered", "pubkey"]:
                val = data.pop(k)
                _unknown[k] = val

        email = data.get("email", None)
        assert email is None or type(email) == str

        registered = data.get("registered", None)
        assert registered is None or type(registered) == float or type(registered) == int
        if registered:
            # registered = datetime.utcfromtimestamp(float(registered) / 1000000.)
            registered = datetime.fromtimestamp(float(registered) / 1000000.0)

        # hex string with 256 bit Ed25519 WAMP-cryptosign public key
        pubkey = data.get("pubkey", None)
        assert pubkey is None or (type(pubkey) == str and len(pubkey) == 64)

        obj = User(
            oid=obj.oid,
            label=obj.label,
            description=obj.description,
            tags=obj.tags,
            email=email,
            registered=registered,
            pubkey=pubkey,
            _unknown=_unknown,
        )
        return obj



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbrmm

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class IPFSFile(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = IPFSFile()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsIPFSFile(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # IPFSFile
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # metadata (IPFS Multihash).
    # IPFSFile
    def FileHash(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Whether the relevant file was downloaded from Infura
    # IPFSFile
    def Downloaded(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Number of download attempts in case of failure
    # IPFSFile
    def Retries(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Timestamp of last error to download
    # IPFSFile
    def ErroredAt(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0


def IPFSFileStart(builder):
    builder.StartObject(4)


def Start(builder):
    IPFSFileStart(builder)


def IPFSFileAddFileHash(builder, fileHash):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(fileHash), 0)


def AddFileHash(builder, fileHash):
    IPFSFileAddFileHash(builder, fileHash)


def IPFSFileAddDownloaded(builder, downloaded):
    builder.PrependBoolSlot(1, downloaded, 0)


def AddDownloaded(builder, downloaded):
    IPFSFileAddDownloaded(builder, downloaded)


def IPFSFileAddRetries(builder, retries):
    builder.PrependUint32Slot(2, retries, 0)


def AddRetries(builder, retries):
    IPFSFileAddRetries(builder, retries)


def IPFSFileAddErroredAt(builder, erroredAt):
    builder.PrependUint64Slot(3, erroredAt, 0)


def AddErroredAt(builder, erroredAt):
    IPFSFileAddErroredAt(builder, erroredAt)


def IPFSFileEnd(builder):
    return builder.EndObject()


def End(builder):
    return IPFSFileEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import uuid
from datetime import datetime

import cbor2
import pytest
import txaio
from zlmdb import flatbuffers

from cfxdb.user import Organization, OrganizationFbs

txaio.use_twisted()


def fill_org(org):
    org.oid = uuid.uuid4()
    org.label = "My Org 1"
    org.description = "Personal organization, created from unit test."
    org.tags = ["previews", "smb"]
    org.name = "homer23"
    org.otype = 1
    org.registered = datetime.utcnow()


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


@pytest.fixture(scope="function")
def org_fbs():
    org = OrganizationFbs()
    fill_org(org)
    return org


@pytest.fixture(scope="function")
def org_cbor():
    org = Organization()
    fill_org(org)
    return org


def test_org_fbs_roundtrip(org_fbs, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = org_fbs.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 192

    # create python object from bytes (flatbuffes)
    _org = OrganizationFbs.cast(data)

    # assert _org == org_fbs

    assert _org.oid == org_fbs.oid
    assert _org.label == org_fbs.label
    assert _org.description == org_fbs.description
    assert _org.tags == org_fbs.tags
    assert _org.name == org_fbs.name
    assert _org.otype == org_fbs.otype
    assert _org.registered == org_fbs.registered


def test_org_copy_fbs_to_cbor(org_fbs):
    org_cbor = Organization()
    org_cbor.copy(org_fbs)

    # assert org_cbor == org_fbs

    assert org_cbor.oid == org_fbs.oid
    assert org_cbor.label == org_fbs.label
    assert org_cbor.description == org_fbs.description
    assert org_cbor.tags == org_fbs.tags
    assert org_cbor.name == org_fbs.name
    assert org_cbor.otype == org_fbs.otype
    assert org_cbor.registered == org_fbs.registered


def test_org_cbor_roundtrip(org_cbor):
    # serialize to bytes (cbor) from python object
    obj = org_cbor.marshal()
    data = cbor2.dumps(obj)
    assert len(data) == 177

    # create python object from bytes (cbor)
    _obj = cbor2.loads(data)
    _org = Organization.parse(_obj)

    # assert _org == org_cbor

    assert _org.oid == org_cbor.oid
    assert _org.label == org_cbor.label
    assert _org.description == org_cbor.description
    assert _org.tags == org_cbor.tags
    assert _org.name == org_cbor.name
    assert _org.otype == org_cbor.otype
    assert _org.registered == org_cbor.registered


def test_org_copy_cbor_to_fbs(org_cbor):
    org_fbs = OrganizationFbs()
    org_fbs.copy(org_cbor)

    # assert org_fbs == org_cbor

    assert org_fbs.oid == org_cbor.oid
    assert org_fbs.label == org_cbor.label
    assert org_fbs.description == org_cbor.description
    assert org_fbs.tags == org_cbor.tags
    assert org_fbs.name == org_cbor.name
    assert org_fbs.otype == org_cbor.otype
    assert org_fbs.registered == org_cbor.registered



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: mrealm

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class ManagementRealm(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ManagementRealm()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsManagementRealm(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # ManagementRealm
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ID of this object.
    # ManagementRealm
    def Oid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Label for this object (not interpreted by CFC).
    # ManagementRealm
    def Label(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Description for this object (not interpreted by CFC).
    # ManagementRealm
    def Description(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Tags on this object.
    # ManagementRealm
    def Tags(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # ManagementRealm
    def TagsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ManagementRealm
    def TagsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # Name of this management realm (must be globally unique within CFC at any given point in time).
    # ManagementRealm
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Time when the management realm was created.
    # ManagementRealm
    def Created(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Owner organization of this management realm.
    # ManagementRealm
    def Owner(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # CFC hosting node for this management realm.
    # ManagementRealm
    def CfNode(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # CFC hosting router worker for this management realm.
    # ManagementRealm
    def CfRouterWorker(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # CFC hosting container worker for this management realm.
    # ManagementRealm
    def CfContainerWorker(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None


def ManagementRealmStart(builder):
    builder.StartObject(10)


def Start(builder):
    ManagementRealmStart(builder)


def ManagementRealmAddOid(builder, oid):
    builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(oid), 0)


def AddOid(builder, oid):
    ManagementRealmAddOid(builder, oid)


def ManagementRealmAddLabel(builder, label):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(label), 0)


def AddLabel(builder, label):
    ManagementRealmAddLabel(builder, label)


def ManagementRealmAddDescription(builder, description):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(description), 0)


def AddDescription(builder, description):
    ManagementRealmAddDescription(builder, description)


def ManagementRealmAddTags(builder, tags):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(tags), 0)


def AddTags(builder, tags):
    ManagementRealmAddTags(builder, tags)


def ManagementRealmStartTagsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartTagsVector(builder, numElems: int) -> int:
    return ManagementRealmStartTagsVector(builder, numElems)


def ManagementRealmAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)


def AddName(builder, name):
    ManagementRealmAddName(builder, name)


def ManagementRealmAddCreated(builder, created):
    builder.PrependUint64Slot(5, created, 0)


def AddCreated(builder, created):
    ManagementRealmAddCreated(builder, created)


def ManagementRealmAddOwner(builder, owner):
    builder.PrependStructSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(owner), 0)


def AddOwner(builder, owner):
    ManagementRealmAddOwner(builder, owner)


def ManagementRealmAddCfNode(builder, cfNode):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(cfNode), 0)


def AddCfNode(builder, cfNode):
    ManagementRealmAddCfNode(builder, cfNode)


def ManagementRealmAddCfRouterWorker(builder, cfRouterWorker):
    builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(cfRouterWorker), 0)


def AddCfRouterWorker(builder, cfRouterWorker):
    ManagementRealmAddCfRouterWorker(builder, cfRouterWorker)


def ManagementRealmAddCfContainerWorker(builder, cfContainerWorker):
    builder.PrependUOffsetTRelativeSlot(9, flatbuffers.number_types.UOffsetTFlags.py_type(cfContainerWorker), 0)


def AddCfContainerWorker(builder, cfContainerWorker):
    ManagementRealmAddCfContainerWorker(builder, cfContainerWorker)


def ManagementRealmEnd(builder):
    return builder.EndObject()


def End(builder):
    return ManagementRealmEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: user

# User activation status (built into code, unconfigurable)
class ActivationStatus(object):
    NONE = 0
    PENDING = 1
    ACTIVE = 2
    EXPIRED = 3



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
import uuid
from typing import Optional

import numpy as np

from cfxdb.mrealm.types import STATUS_BY_CODE, STATUS_BY_NAME


class RouterWorkerGroupClusterPlacement(object):
    """
    Placement of router worker groups onto router clusters, specifically **router workers**
    and **xbrmm workers** running as part of router worker groups.
    """

    def __init__(
        self,
        oid: Optional[uuid.UUID] = None,
        worker_group_oid: Optional[uuid.UUID] = None,
        cluster_oid: Optional[uuid.UUID] = None,
        node_oid: Optional[uuid.UUID] = None,
        worker_name: Optional[str] = None,
        status: Optional[int] = None,
        changed: Optional[np.datetime64] = None,
        tcp_listening_port: Optional[int] = None,
        _unknown: Optional[dict] = None,
    ):
        """

        :param oid: Object ID of this placement itself.

        :param worker_group_oid: Object ID of the router worker group this placement applies to.
            Refers to :class:`cfxdb.mrealm.RouterWorkerGroup`

        :param cluster_oid: Object ID of the router cluster this placement applies to.
            Refers to :class:`cfxdb.mrealm.RouterCluster`

        :param node_oid: Object ID of the node (within the router cluster) this placement is assigned to.
            Refers to :class:`cfxdb.mrealm.Node`

        :param worker_name: Run-time ID (in the node) of the router worker this placement is assigned to.

        :param status: Status of this placement, which essentially reflects the router worker status of this placement.

        :param changed: Timestamp when the status of this placement last changed.

        :param tcp_listening_port: TCP listening port the router worker this placement is assigned to is listening on
            for incoming proxy front-end and router-to-router connections.
        """
        self.oid = oid
        self.worker_group_oid = worker_group_oid
        self.cluster_oid = cluster_oid
        self.node_oid = node_oid
        self.worker_name = worker_name
        self.status = status
        self.changed = changed
        self.tcp_listening_port = tcp_listening_port
        self._unknown = _unknown

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if other.oid != self.oid:
            return False
        if other.worker_group_oid != self.worker_group_oid:
            return False
        if other.cluster_oid != self.cluster_oid:
            return False
        if other.node_oid != self.node_oid:
            return False
        if other.worker_name != self.worker_name:
            return False
        if other.status != self.status:
            return False
        if other.changed != self.changed:
            return False
        if other.tcp_listening_port != self.tcp_listening_port:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return pprint.pformat(self.marshal())

    def marshal(self):
        """
        Marshal this object to a generic host language object.

        :return: dict
        """
        obj = {
            "oid": str(self.oid) if self.oid else None,
            "worker_group_oid": str(self.worker_group_oid),
            "cluster_oid": str(self.cluster_oid),
            "node_oid": str(self.node_oid),
            "worker_name": self.worker_name,
            "status": STATUS_BY_CODE[self.status] if self.status else None,
            "changed": int(self.changed) if self.changed else None,
            "tcp_listening_port": self.tcp_listening_port,
        }
        return obj

    @staticmethod
    def parse(data):
        """
        Parse generic host language object into an object of this class.

        :param data: Generic host language object
        :type data: dict

        :return: instance of :class:`WebService`
        """
        assert type(data) == dict

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = {}
        for k in data:
            if k not in [
                "oid",
                "worker_group_oid",
                "cluster_oid",
                "node_oid",
                "worker_name",
                "status",
                "changed",
                "tcp_listening_port",
            ]:
                _unknown[k] = data[k]

        oid = None
        if "oid" in data:
            assert type(data["oid"]) == str
            oid = uuid.UUID(data["oid"])

        worker_group_oid = None
        if "worker_group_oid" in data:
            assert type(data["worker_group_oid"]) == str
            worker_group_oid = uuid.UUID(data["worker_group_oid"])

        cluster_oid = None
        if "cluster_oid" in data:
            assert type(data["cluster_oid"]) == str
            cluster_oid = uuid.UUID(data["cluster_oid"])

        node_oid = None
        if "node_oid" in data:
            assert type(data["node_oid"]) == str
            node_oid = uuid.UUID(data["node_oid"])

        worker_name = None
        if "worker_name" in data:
            assert type(data["worker_name"]) == str
            worker_name = data["worker_name"]

        status = data.get("status", None)
        assert status is None or (type(status) == str)
        status = STATUS_BY_NAME.get(status, None)

        changed = data.get("changed", None)
        assert changed is None or (type(changed) == int)
        if changed:
            changed = np.datetime64(changed, "ns")

        tcp_listening_port = data.get("tcp_listening_port", None)
        assert tcp_listening_port is None or (type(tcp_listening_port) == int and 0 <= tcp_listening_port < 65536)

        obj = RouterWorkerGroupClusterPlacement(
            oid=oid,
            worker_group_oid=worker_group_oid,
            cluster_oid=cluster_oid,
            worker_name=worker_name,
            node_oid=node_oid,
            status=status,
            changed=changed,
            tcp_listening_port=tcp_listening_port,
            _unknown=_unknown,
        )

        return obj



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
import uuid

import numpy as np
from zlmdb import MapBytes20FlatBuffers, flatbuffers, table

from cfxdb import pack_uint256, unpack_uint256
from cfxdb.gen.xbr import Member as MemberGen
from cfxdb.gen.xbr.MemberLevel import MemberLevel


class _MemberGen(MemberGen.Member):
    """
    Expand methods on the class code generated by flatc.

    FIXME: come up with a PR for flatc to generated this stuff automatically.
    """

    @classmethod
    def GetRootAsMember(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _MemberGen()
        x.Init(buf, n + offset)
        return x

    def AddressAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def AccountOidAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def RegisteredAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def TidAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def SignatureAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class Member(object):
    """
    XBR Network member database object.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        # [uint8] (address)
        self._address = None

        # [uint8] (uuid)
        self._account_oid = None

        # uint64 (timestamp)
        self._timestamp = None

        # [uint8] (uint256)
        self._registered = None

        # string (multihash)
        self._eula = None

        # string (multihash)
        self._profile = None

        # enum MemberLevel: uint8
        self._level = None

        # [uint8] (ethhash)
        self._tid = None

        # [uint8] (ethsig)
        self._signature = None

    def marshal(self) -> dict:
        obj = {
            "address": bytes(self.address) if self._address else None,
            "account_oid": self._account_oid.bytes if self._account_oid else None,
            "timestamp": int(self.timestamp) if self.timestamp else None,
            "registered": self.registered,
            "eula": self.eula,
            "profile": self.profile,
            "level": self.level,
            "tid": bytes(self.tid) if self.tid else None,
            "signature": bytes(self.signature) if self.signature else None,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def address(self) -> bytes:
        """
        Ethereum address of the member.
        """
        if self._address is None and self._from_fbs:
            if self._from_fbs.AddressLength():
                self._address = self._from_fbs.AddressAsBytes()
        return self._address

    @address.setter
    def address(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 20)
        self._address = value

    @property
    def account_oid(self) -> uuid.UUID:
        """
        ID of user account this member has an account on planet.xbr.network.
        """
        if self._account_oid is None and self._from_fbs:
            if self._from_fbs.AccountOidLength():
                _account_oid = self._from_fbs.AccountOidAsBytes()
                self._account_oid = uuid.UUID(bytes=bytes(_account_oid))
            else:
                self._account_oid = uuid.UUID(bytes=b"\x00" * 16)
        return self._account_oid

    @account_oid.setter
    def account_oid(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._account_oid = value

    @property
    def timestamp(self) -> np.datetime64:
        """
        Database transaction time (epoch time in ns) of insert or last update.
        """
        if self._timestamp is None and self._from_fbs:
            self._timestamp = np.datetime64(self._from_fbs.Timestamp(), "ns")
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._timestamp = value

    @property
    def registered(self) -> int:
        """
        Block number (on the blockchain) when the member (originally) registered.
        """
        if self._registered is None and self._from_fbs:
            if self._from_fbs.RegisteredLength():
                _registered = self._from_fbs.RegisteredAsBytes()
                self._registered = unpack_uint256(bytes(_registered))
            else:
                self._registered = 0
        return self._registered

    @registered.setter
    def registered(self, value: int):
        assert value is None or type(value) == int
        self._registered = value

    @property
    def eula(self) -> str:
        """
        EULA the member agreed to when joining the market (IPFS Multihash string).
        """
        if self._eula is None and self._from_fbs:
            eula = self._from_fbs.Eula()
            if eula:
                self._eula = eula.decode("utf8")
        return self._eula

    @eula.setter
    def eula(self, value):
        assert value is None or type(value) == str
        self._eula = value

    @property
    def profile(self) -> str:
        """
        Optional member profile (IPFS Multihash string).
        """
        if self._profile is None and self._from_fbs:
            profile = self._from_fbs.Profile()
            if profile:
                self._profile = profile.decode("utf8")
        return self._profile

    @profile.setter
    def profile(self, value):
        assert value is None or type(value) == str
        self._profile = value

    @property
    def level(self) -> int:
        """
        Current member level.
        """
        if self._level is None and self._from_fbs:
            self._level = self._from_fbs.Level()
        return self._level

    @level.setter
    def level(self, value: int):
        assert value is None or type(value) == int
        assert value in [
            MemberLevel.NONE,
            MemberLevel.ACTIVE,
            MemberLevel.VERIFIED,
            MemberLevel.RETIRED,
            MemberLevel.PENALTY,
            MemberLevel.BLOCKED,
        ]
        self._level = value

    @property
    def tid(self) -> bytes:
        """
        Transaction hash of the transaction this change was committed to the blockchain under.
        """
        if self._tid is None and self._from_fbs:
            if self._from_fbs.TidLength():
                self._tid = self._from_fbs.TidAsBytes()
        return self._tid

    @tid.setter
    def tid(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 32)
        self._tid = value

    @property
    def signature(self) -> bytes:
        """
        When signed off-chain and submitted via ``XBRNetwork.registerMemberFor``.
        """
        if self._signature is None and self._from_fbs:
            if self._from_fbs.SignatureLength():
                self._signature = self._from_fbs.SignatureAsBytes()
        return self._signature

    @signature.setter
    def signature(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 65)
        self._signature = value

    @staticmethod
    def cast(buf):
        return Member(_MemberGen.GetRootAsMember(buf, 0))

    def build(self, builder):
        address = self.address
        if address:
            address = builder.CreateString(address)

        account_oid = self.account_oid
        if account_oid:
            account_oid = builder.CreateString(account_oid.bytes)

        registered = self.registered
        if registered:
            registered = builder.CreateString(pack_uint256(registered))

        eula = self.eula
        if eula:
            eula = builder.CreateString(eula)

        profile = self.profile
        if profile:
            profile = builder.CreateString(profile)

        tid = self.tid
        if tid:
            tid = builder.CreateString(tid)

        signature = self.signature
        if signature:
            signature = builder.CreateString(signature)

        MemberGen.MemberStart(builder)

        if address:
            MemberGen.MemberAddAddress(builder, address)

        if account_oid:
            MemberGen.MemberAddAccountOid(builder, account_oid)

        if self.timestamp:
            MemberGen.MemberAddTimestamp(builder, int(self.timestamp))

        if registered:
            MemberGen.MemberAddRegistered(builder, registered)

        if eula:
            MemberGen.MemberAddEula(builder, eula)

        if profile:
            MemberGen.MemberAddProfile(builder, profile)

        if self.level:
            MemberGen.MemberAddLevel(builder, self.level)

        if tid:
            MemberGen.MemberAddTid(builder, tid)

        if signature:
            MemberGen.MemberAddSignature(builder, signature)

        final = MemberGen.MemberEnd(builder)

        return final


@table("d1808139-5a3b-4a4e-abad-152dd4cd1131", build=Member.build, cast=Member.cast)
class Members(MapBytes20FlatBuffers):
    """
    XBR members by ``member_adr``.

    Map :class:`zlmdb.MapBytes20FlatBuffers` from ``member_adr`` to :class:`cfxdb.xbr.Member`
    """



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: realmstore

# The payload end-to-end encryption algorithm the payload is encrypted.
class EncAlgo(object):
    # Unset
    NONE = 0
    # WAMP-cryptobox end-to-end encrypted application payload
    CRYPTOBOX = 1
    # MQTT passthrough-mode application payload
    MQTT = 2
    # XBR end-to-end encrypted application payload
    XBR = 3
    OPAQUE = 4



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class ObjRef(object):
    __slots__ = ["_tab"]

    @classmethod
    def SizeOf(cls):
        return 24

    # ObjRef
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # OID of the referenced object: 128 bit with UUID.
    # ObjRef
    def Oid(self, obj):
        obj.Init(self._tab.Bytes, self._tab.Pos + 0)
        return obj

    # Version of the referenced object, starting with 1, and incremented
    # ObjRef
    def Version(self):
        return self._tab.Get(
            flatbuffers.number_types.Uint32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(16)
        )


def CreateObjRef(builder, oid_msb, oid_lsb, version):
    builder.Prep(8, 24)
    builder.Pad(4)
    builder.PrependUint32(version)
    builder.Prep(8, 16)
    builder.PrependUint64(oid_lsb)
    builder.PrependUint64(oid_msb)
    return builder.Offset()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: arealm

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# Application realm client principals.
class Principal(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Principal()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsPrincipal(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Principal
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ID of this principal.
    # Principal
    def Oid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Timestamp when the principal was last modified
    # Principal
    def Modified(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # ID of the application realm the authenticated principal will be joined to.
    # Principal
    def ArealmOid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # WAMP authid of the principal, must be unique within the application realm at any moment in time.
    # Principal
    def Authid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ID of the role the authenticated principal will be joined to the application realm.
    # Principal
    def RoleOid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Optional authextra information returned to the authenticating principal (CBOR serialized).
    # Principal
    def Authextra(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Principal
    def AuthextraAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Principal
    def AuthextraLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Principal
    def AuthextraIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0


def PrincipalStart(builder):
    builder.StartObject(6)


def Start(builder):
    PrincipalStart(builder)


def PrincipalAddOid(builder, oid):
    builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(oid), 0)


def AddOid(builder, oid):
    PrincipalAddOid(builder, oid)


def PrincipalAddModified(builder, modified):
    builder.PrependInt32Slot(1, modified, 0)


def AddModified(builder, modified):
    PrincipalAddModified(builder, modified)


def PrincipalAddArealmOid(builder, arealmOid):
    builder.PrependStructSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(arealmOid), 0)


def AddArealmOid(builder, arealmOid):
    PrincipalAddArealmOid(builder, arealmOid)


def PrincipalAddAuthid(builder, authid):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(authid), 0)


def AddAuthid(builder, authid):
    PrincipalAddAuthid(builder, authid)


def PrincipalAddRoleOid(builder, roleOid):
    builder.PrependStructSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(roleOid), 0)


def AddRoleOid(builder, roleOid):
    PrincipalAddRoleOid(builder, roleOid)


def PrincipalAddAuthextra(builder, authextra):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(authextra), 0)


def AddAuthextra(builder, authextra):
    PrincipalAddAuthextra(builder, authextra)


def PrincipalStartAuthextraVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartAuthextraVector(builder, numElems: int) -> int:
    return PrincipalStartAuthextraVector(builder, numElems)


def PrincipalEnd(builder):
    return builder.EndObject()


def End(builder):
    return PrincipalEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbr

# XBR Member level.
class MemberLevel(object):
    # Unset
    NONE = 0
    # Member is active.
    ACTIVE = 1
    # Member is active and verified.
    VERIFIED = 2
    # Member is retired.
    RETIRED = 3
    # Member is subject to a temporary penalty.
    PENALTY = 4
    # Member is currently blocked and cannot current actively participate in the market.
    BLOCKED = 5



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: arealm

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# WAMP client authentication credentials, used for mapping ``(authmethod, realm, authid) -> principal``.
class Credential(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Credential()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsCredential(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Credential
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ID of this credential.
    # Credential
    def Oid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # WAMP authentication method offered by the authenticating client.
    # Credential
    def Authmethod(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # WAMP realm requested by the authenticating client.
    # Credential
    def Realm(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # WAMP authid announced by the authenticating client.
    # Credential
    def Authid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Authentication method specific configuration (CBOR serialized).
    # Credential
    def Authconfig(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Credential
    def AuthconfigAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Credential
    def AuthconfigLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Credential
    def AuthconfigIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # ID of the principal this credential resolves to upon successful authentication.
    # Credential
    def PrincipalOid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None


def CredentialStart(builder):
    builder.StartObject(6)


def Start(builder):
    CredentialStart(builder)


def CredentialAddOid(builder, oid):
    builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(oid), 0)


def AddOid(builder, oid):
    CredentialAddOid(builder, oid)


def CredentialAddAuthmethod(builder, authmethod):
    builder.PrependInt8Slot(1, authmethod, 0)


def AddAuthmethod(builder, authmethod):
    CredentialAddAuthmethod(builder, authmethod)


def CredentialAddRealm(builder, realm):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(realm), 0)


def AddRealm(builder, realm):
    CredentialAddRealm(builder, realm)


def CredentialAddAuthid(builder, authid):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(authid), 0)


def AddAuthid(builder, authid):
    CredentialAddAuthid(builder, authid)


def CredentialAddAuthconfig(builder, authconfig):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(authconfig), 0)


def AddAuthconfig(builder, authconfig):
    CredentialAddAuthconfig(builder, authconfig)


def CredentialStartAuthconfigVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartAuthconfigVector(builder, numElems: int) -> int:
    return CredentialStartAuthconfigVector(builder, numElems)


def CredentialAddPrincipalOid(builder, principalOid):
    builder.PrependStructSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(principalOid), 0)


def AddPrincipalOid(builder, principalOid):
    CredentialAddPrincipalOid(builder, principalOid)


def CredentialEnd(builder):
    return builder.EndObject()


def End(builder):
    return CredentialEnd(builder)
